/*------------------------------------------------------------------------------------------*\
 *   
 *   Copyright (C) 2011 AVM GmbH <fritzbox_info@avm.de>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 *   
 *
 *   Handle unaligned accesses in mips16 mode by emulation.
\*------------------------------------------------------------------------------------------*/


#include <linux/mm.h>
#include <linux/module.h>
#include <linux/signal.h>
#include <linux/smp.h>
#include <linux/sched.h>
#include <linux/debugfs.h>
#include <asm/asm.h>
#include <asm/branch.h>
#include <asm/byteorder.h>
#include <asm/inst.h>
#include <asm/inst_mips16.h>
#include <asm/uaccess.h>
#include <asm/system.h>

#define REG_SP 29
#define REG_RA 31

#ifdef CONFIG_DEBUG_FS
extern u32 unaligned_instructions;
extern u32 unaligned_action;
#else
#define unaligned_action UNALIGNED_ACTION_QUIET
#endif

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void emulate_load_store_insn_mips16(struct pt_regs *regs,
	void __user *addr, unsigned int __user *pc)
{
	union mips16_instruction insn;

#ifdef __BIG_ENDIAN
    short offset, rx, ry;
    unsigned char *base, *val;
    unsigned int dest = 0;
#endif

	regs->regs[0] = 0;

    if(read_mips16_insn(&insn, (unsigned int)pc & ~0x1 ))
        goto sigsegv;

    switch (insn.meta.opcode) {
#ifdef __BIG_ENDIAN
        case mips16_op_lwsp: /*--- load word SP-relative ---*/
        case mips16_op_lwpc: /*--- load word PC-relative ---*/
            get_mips16_rx(&insn, &rx, &offset, 2);
            if(insn.meta.opcode == mips16_op_lwsp)
                base = (unsigned char *)regs->regs[REG_SP];
            else
                base = (unsigned char *)regs->cp0_epc; /*--- in delay slot: addr of jump instruction ---*/

            printk(KERN_ERR "[%s] LW%s%s r%d, 0x%x ( %s ) #   0x%lx = *(unsigned int *)(0x%x + 0x%x)\n", __FUNCTION__,  
                    (insn.meta.opcode == mips16_op_lwsp) ? "SP" : "PC",
                    insn.meta.is_extended ? "_EXT" : "", rx,  offset, 
                    (insn.meta.opcode == mips16_op_lwsp) ? "SP" : "PC",
                    regs->regs[rx], (unsigned int)base,  offset);

            dest |= base[offset + 0] << 24;
            dest |= base[offset + 1] << 16;
            dest |= base[offset + 2] <<  8;
            dest |= base[offset + 3] <<  0;

            if((insn.meta.opcode == mips16_op_lwpc) && (((int)base+offset) & 0x11)) {
                /*--- => XXX auch wenn (!insn.meta.is_extended)? Oder dann nur unteren 2 Bits löschen? ---*/
                goto sigbus;
            }

            regs->regs[rx] = dest;

            compute_return_epc(regs);
            break;

        case mips16_op_lw:
            get_mips16_rxy(&insn, &rx, &ry, &offset, 2);
            base  = (unsigned char *)(regs->regs[rx]);

            printk(KERN_ERR "[%s] LW%s r%d, 0x%x ( r%d ) #   0x%lx = *(unsigned int *)(0x%x + 0x%x)\n", __FUNCTION__,  
                    insn.meta.is_extended ? "_EXT" : "", ry, offset, rx, regs->regs[ry], (unsigned int)base,  offset);

            dest |= base[offset + 0] << 24;
            dest |= base[offset + 1] << 16;
            dest |= base[offset + 2] <<  8;
            dest |= base[offset + 3] <<  0;
            regs->regs[ry] = dest;
    
            compute_return_epc(regs);
            break;

        case mips16_op_lhu:
        case mips16_op_lh:
            get_mips16_rxy(&insn, &rx, &ry, &offset, 1);
            base  = (unsigned char *)(regs->regs[rx]);

            printk(KERN_ERR "[%s] LH%s r%d, 0x%x ( r%d ) #   0x%lx = *(unsigned int *)(0x%x + 0x%x)\n", __FUNCTION__,  
                    insn.meta.is_extended ? "_EXT" : "", ry, offset, rx, regs->regs[ry], (unsigned int)base,  offset);

            dest |= base[offset + 0] << 8;
            dest |= base[offset + 1] << 0;
            regs->regs[ry] = (insn.meta.opcode == mips16_op_lh) ? (short)dest : (unsigned short)dest;
            compute_return_epc(regs);
            break;

        case mips16_op_sh:
            get_mips16_rxy(&insn, &rx, &ry, &offset, 1);
            base = (unsigned char *)(regs->regs[rx]);
            val  = (unsigned char *)&(regs->regs[ry]);

            printk(KERN_ERR "[%s] SH%s r%d, 0x%x ( r%d ) #   0x%lx = *(unsigned int *)(0x%x + 0x%x)\n", __FUNCTION__,  
                    insn.meta.is_extended ? "_EXT" : "", ry, offset, rx, regs->regs[ry], (unsigned int)base,  offset);

            base[offset + 0] = val[0];
            base[offset + 1] = val[1];
            compute_return_epc(regs);
            break;

        case mips16_op_sw:
            get_mips16_rxy(&insn, &rx, &ry, &offset, 2);
            base = (unsigned char *)(regs->regs[rx]);
            val  = (unsigned char *)&(regs->regs[ry]);

            printk(KERN_ERR "[%s] SW%s r%d, 0x%x ( r%d ) #   0x%lx = *(unsigned int *)(0x%x + 0x%x)\n", __FUNCTION__,  
                    insn.meta.is_extended ? "_EXT" : "", ry, offset, rx, regs->regs[ry], (unsigned int)base,  offset);

            base[offset + 0] = val[0];
            base[offset + 1] = val[1];
            base[offset + 2] = val[2];
            base[offset + 3] = val[3];
            compute_return_epc(regs);
            break;

        case mips16_op_i8: 
            if((insn.meta.is_extended && (insn.ext_rr_format.rx != mips16_i8_swrasp)) ||
                    (insn.rr_format.rx != mips16_i8_swrasp)) {
                goto sigbus;
            }

        case mips16_op_swsp: /*--- store word SP-relative ---*/
            get_mips16_rx(&insn, &rx, &offset, 2);
            if(insn.meta.opcode == mips16_op_lwsp)
                rx = regs->regs[REG_RA]; /*--- SWRASP: store RA-Register SP-relatve---*/ 
            base = (unsigned char *)(regs->regs[REG_SP]);
            val  = (unsigned char *)&(regs->regs[rx]);

            printk(KERN_ERR "[%s] SWSP%s r%d, 0x%x ( SP ) #   0x%lx = *(unsigned int *)(0x%x + 0x%x)\n", __FUNCTION__,  
                    insn.meta.is_extended ? "_EXT" : "", rx, offset, regs->regs[rx], (unsigned int)base,  offset);

            base[offset + 0] = val[0];
            base[offset + 1] = val[1];
            base[offset + 2] = val[2];
            base[offset + 3] = val[3];
            compute_return_epc(regs);
            break;

#endif /*--- #ifdef __BIG_ENDIAN ---*/
        default: goto sigbus;
    }

#ifdef CONFIG_DEBUG_FS
	unaligned_instructions++;
#endif

	return;

/*--- fault: ---*/
	/*--- Did we have an exception handler installed? ---*/
	/*--- if (fixup_exception(regs)) ---*/
		/*--- return; ---*/

	/*--- die_if_kernel("Unhandled kernel unaligned access", regs); ---*/
	/*--- send_sig(SIGSEGV, current, 1); ---*/

	/*--- return; ---*/

sigbus:
    printk(KERN_ERR "[%s] ERROR - no unaligned handling for instruction 0x%x (opcode=%d%s) at address 0x%x\n",
            __FUNCTION__,  insn.word, insn.meta.opcode, insn.meta.is_extended ? ", extended" : "", (unsigned int)pc);
	die_if_kernel("Unhandled kernel unaligned access", regs);
	send_sig(SIGBUS, current, 1);

	return;

sigsegv:
	printk("[%s] %s: ERROR on decoding instruction - sending SIGSEGV.\n", __FUNCTION__, current->comm);
	force_sig(SIGSEGV, current);
	return;

/*--- sigill: ---*/
	/*--- die_if_kernel("Unhandled kernel unaligned access or invalid instruction", regs); ---*/
	/*--- send_sig(SIGILL, current, 1); ---*/

}


