/*------------------------------------------------------------------------------------------*\
 *   Copyright (C) 2006,2007, ... ,2011 AVM GmbH <fritzbox_info@avm.de>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
\*------------------------------------------------------------------------------------------*/


#if defined(CONFIG_AVM_LED_EVENTS) || defined(CONFIG_AVM_LED_EVENTS_MODULE)
#include <linux/avm_led_event.h>
#endif /*--- #if defined(CONFIG_AVM_LED_EVENTS) || defined(CONFIG_AVM_LED_EVENTS_MODULE) ---*/

#include "cpmac_debug.h"
#include "cpmac_interface.h"
#include "cpphy_types.h"
#include "cpphy_mgmt.h"
#include <ifx_gpio.h>

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void cpmac_interface_check_link_dummy(phy_number_t phy_number __attribute__ ((unused))) {
    /* dummy function */
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
unsigned int cpmac_interface_print_link_state_output(unsigned char *output, unsigned int length) {
    unsigned char port;
    unsigned int written = 0;

    if(length < 200) { /* Rough upper bound for output size */
        DEB_WARN("[%s] Buffer length (%u) too short for output.\n", __FUNCTION__, length);
        return 0;
    }
    written += snprintf(output + written, length - written, "Port stati: ");
    for(port = 0; port < cpmac_global.ports; port++) {
        if(cpmac_global.event_data.port[port].link) {
            written += snprintf(output + written,
                                CPMAC_DEBUG_BUFFER_LENGTH - written,
                                "[%u - %s%s]  ", 
                                port,
                                (cpmac_global.event_data.port[port].speed == avm_event_ethernet_speed_1G) 
                                ? " 1000" 
                                : ((cpmac_global.event_data.port[port].speed == avm_event_ethernet_speed_100M) 
                                ? " 100" 
                                : ((cpmac_global.event_data.port[port].speed == avm_event_ethernet_speed_10M) 
                                ? " 10" : " speed?")),
                                cpmac_global.event_data.port[port].fullduplex ? " FD" : " HD");
        } else {
            written += snprintf(output + written,
                                CPMAC_DEBUG_BUFFER_LENGTH - written,
                                "[%u - down]  ", port);
        }
    }
    written += snprintf(output + written, CPMAC_DEBUG_BUFFER_LENGTH - written, "\n");
    return written;
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void cpmac_interface_print_link_state(void) {
#   if(CPMAC_DEBUG_LEVEL & CPMAC_DEBUG_LEVEL_TRACE)
    unsigned char *output;
    
    output = kmalloc(CPMAC_DEBUG_BUFFER_LENGTH + 1, GFP_KERNEL);
    if(output == NULL) {
        DEB_WARN("[%s] out of memory\n", __FUNCTION__);
        return;
    }
    cpmac_interface_print_link_state_output(output, CPMAC_DEBUG_BUFFER_LENGTH);
    DEB_TRC("%s", output);
    kfree(output);
#   endif /*--- #if(CPMAC_DEBUG_LEVEL & CPMAC_DEBUG_LEVEL_TRACE) ---*/
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void cpmac_interface_port_update(unsigned char                  port, 
                                 unsigned char                  link,
                                 enum _avm_event_ethernet_speed speed,
                                 unsigned char                  fullduplex) {
    cpmac_phy_t *phy;
    enum { LINK_NOT_CHANGED = 0, LINK_IS_UP, LINK_IS_DOWN } link_change = LINK_NOT_CHANGED;

    if(port >= cpmac_global.ports) {
        DEB_ERR("[%s] Maximum port number %u exceeded with %u!\n", 
                __FUNCTION__, cpmac_global.ports, port);
    }
    phy = &cpmac_global.phy[cpmac_global.map_port_to_phy[port]];

    if(   (cpmac_global.event_data.port[port].link == link)
       && (cpmac_global.event_data.port[port].speed == speed)
       && (cpmac_global.event_data.port[port].fullduplex == fullduplex)
       && (speed != avm_event_ethernet_speed_error)) {
        /* No link details changed */
        DEB_DEBUG("[%s] Got port %u update without changes\n", __FUNCTION__, port);
        return;
    }

    if (speed == avm_event_ethernet_speed_error) {
        link_change = LINK_IS_DOWN;
        speed = avm_event_ethernet_speed_no_link;
    }

    if(phy->f->set_port_speed) {
        phy->f->set_port_speed(port, speed, fullduplex);
        DEB_TRC("[%s] Set hw port %u speed to %s %s\n", 
                __FUNCTION__, 
                port, 
                cpmac_global.lan_speed_names[speed],
                fullduplex ? "FD" : "HD");
    }

    if(cpmac_global.event_data.port[port].link != link) {
        if(link) {
            /* Link came up */
            DEB_INFO("[%s] LINK is UP: Port Number %u\n", __FUNCTION__, port);
            link_change = LINK_IS_UP;
        } else {
            /* Link went down */
            DEB_INFO("[%s] LINK is DOWN: Port Number %u\n", __FUNCTION__, port);
            link_change = LINK_IS_DOWN;
        }
    }

    cpmac_global.event_data.port[port].link = link;
    cpmac_global.event_data.port[port].speed = speed;
    cpmac_global.event_data.port[port].fullduplex = fullduplex;

    cpphy_mgmt_link_changed();

    if(link_change == LINK_IS_DOWN) {
        DEB_INFO("[%s] LINK change: Port Number %u\n", __FUNCTION__, port);
        if(port >= 2) {
            DEB_INFO("[%s] reset PHY %u\n", __FUNCTION__, port);
            cpphy_mgmt_reset_port(port, 1);
            msleep(500);
            cpphy_mgmt_reset_port(port, 0);
            if(phy->f->init_phy) {
                phy->f->init_phy(port);
            }
        }
    }
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void cpmac_interface_event_update(void) {
    struct cpmac_event_struct *event_struct;

    if(!avm_event_source_check_id(cpmac_global.event_handle, avm_event_id_ethernet_connect_status)) {
        DEB_TRC("[%s] no one interested in event\n", __FUNCTION__);
        return;
    }
    if(!(event_struct = (struct cpmac_event_struct *) kmalloc(sizeof(struct cpmac_event_struct), GFP_ATOMIC))) {
        DEB_ERR("[%s] No memory for status update!\n", __FUNCTION__);
        return;
    }

    memcpy((unsigned char *) event_struct,
           (unsigned char *) &cpmac_global.event_data,
           sizeof (struct cpmac_event_struct));
    event_struct->event_header.id = avm_event_id_ethernet_connect_status;
    avm_event_source_trigger(cpmac_global.event_handle,
                             avm_event_id_ethernet_connect_status,
                             sizeof(struct cpmac_event_struct),
                             (void *) event_struct);
}


/*------------------------------------------------------------------------------------------*\
  Event update requested from application
\*------------------------------------------------------------------------------------------*/
static void cpmac_interface_event_notify(void *context __attribute__ ((unused)), 
                                         enum _avm_event_id id) {
    if(id == avm_event_id_ethernet_connect_status) {
        cpmac_interface_event_update();
    } else {
        DEB_WARN("[%s] unhandled id %u!\n", __FUNCTION__, id);
    }
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
cpmac_err_t cpmac_interface_event_init(void) {
    cpmac_global.event_data.ports = cpmac_global.ports;
    cpmac_global.event_handle = avm_event_source_register("Ethernet status",
                                                          /*--- (1ull << avm_event_id_ethernet_status) ---*/
                                                          (1ull << avm_event_id_ethernet_connect_status),
                                                          cpmac_interface_event_notify, NULL);
    if(!cpmac_global.event_handle) {
        DEB_ERR("[%s] Could not register avm_event source!\n", __FUNCTION__);
        return CPMAC_ERR_REGISTER_FAILED;
    }
    return CPMAC_ERR_NOERR;
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void cpmac_interface_event_deinit(void) {
    avm_event_source_release(cpmac_global.event_handle);
    cpmac_global.event_handle = NULL;
}


