/*------------------------------------------------------------------------------------------*\
 *   Copyright (C) 2006,2007, ... ,2011 AVM GmbH <fritzbox_info@avm.de>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
\*------------------------------------------------------------------------------------------*/


#include <stdarg.h>
#include <linux/interrupt.h>
#include <linux/timer.h>
#include <linux/skbuff.h>
#include <linux/err.h>
#include <linux/delay.h>
#include <linux/netdevice.h>
#include <net/pkt_sched.h>

#include <linux/avm_cpmac.h>

#include "cpgmac_f.h"
#include "cpmac_if.h"
#include "cpmac_const.h"
#include "cpmac_debug.h"
#include "cpmac_main.h"
#include "cpphy_const.h"
#include "cpphy_types.h"        /* for cpphy_global_t */
#include "cpphy_if.h"           /* for cpphy_if_tx_complete */
#include "cpphy_adm6996.h"      /* for ADM_GET_TAG_GROUP */
#include "cpphy_ar8216.h"       /* for ATH_GET_TAG_GROUP */
#include "cpphy_mgmt.h"         /* for cpphy_mdio_event_dataupdate */
#include "cpphy_mdio.h"         /* for cpphy_mdio_event_update */
#include "cpphy_cppi.h"         /* for cpphy_cppi_teardown */
#include "cpmac_puma_if.h"
#include <asm/mach_avm.h>
#include "cpmac_multicast.h"

#include <linux/if_vlan.h>


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static void cpmac_mcfw_handle_tasklet(unsigned long data) {
    struct sk_buff *skb;
    int i = 32;
    unsigned char instance_number = (unsigned char) (data & 0xff);
    cpmac_instance_t *instance;

    assert(instance_number < AVM_CPMAC_MAX_INSTANCES);
    instance = &cpmac_global.instance[instance_number];

    while(i-- > 0 && (skb = skb_dequeue(&instance->mcfw_queue)) != 0)
        dev_queue_xmit(skb);
    if(skb_queue_len(&instance->mcfw_queue))
        tasklet_schedule(&instance->mcfw_tasklet);
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static void cpmac_dev_queue_xmit(struct net_device *dev, 
                                 struct sk_buff    *skb,
                                 unsigned char      instance_number) {
    cpmac_instance_t *instance;
    struct ethhdr *eth = (struct ethhdr *) skb->data;

    assert(instance_number < AVM_CPMAC_MAX_INSTANCES);
    instance = &cpmac_global.instance[instance_number];

    memcpy(eth->h_source, dev->dev_addr, ETH_ALEN);
    skb->dev = dev;

    if(irqs_disabled()) {       /* tiatm is source */
        skb_queue_tail(&instance->mcfw_queue, skb);
        tasklet_schedule(&instance->mcfw_tasklet);
    } else {
        dev_queue_xmit(skb);
    }
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static void cpmac_mc_transmit(void           *privatedata,
                              int             sourceid __attribute__ ((unused)),
                              mcfw_portset    portset,
                              struct sk_buff *skb) {
    unsigned char instance_number = (unsigned char) (((unsigned int) privatedata) & 0xff);
    cpmac_instance_t *instance;
    struct net_device *p_dev;
    struct sk_buff *tskb;
    unsigned short vid;
    t_cpphy_switch_config *switch_config;

    assert(instance_number < AVM_CPMAC_MAX_INSTANCES);
    instance = &cpmac_global.instance[instance_number];

#   if defined(CONFIG_ARCH_PUMA5)
    p_dev = cpmac_puma_get_dev();
#   else /*--- #if defined(CONFIG_ARCH_PUMA5) ---*/
    p_dev = instance->dev;
    switch_config = &instance->mdio.switch_config;
#   endif /*--- #else ---*/ /*--- #if defined(CONFIG_ARCH_PUMA5) ---*/

    if(p_dev == 0) {
        dev_kfree_skb_any(skb);
        return;
    }

    /* MC streams go to the LAN */
    skb->uniq_id &= 0xffffff;
    skb->uniq_id |= (CPPHY_PRIO_QUEUE_LAN << 24);
    assert(portset <= AVM_CPMAC_MAX_PORTSET);
    /*--- vid = switch_config->map_portset_out[portset]; ---*/
    vid = cpmac_global.instance[0].mdio.switch_config.map_portset_out[portset]; /* FIXME */
    if(vid != 0) {
        if((tskb = __vlan_put_tag(skb, vid)) != 0) {
            cpmac_dev_queue_xmit(p_dev, tskb, instance_number);
        } else {
            dev_kfree_skb_any(skb);
        }
    } else {
        cpmac_dev_queue_xmit(p_dev, skb, instance_number);
    }
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static void cpmac_mc_transmit_single(void            *privatedata,
                                     int              sourceid __attribute__ ((unused)),
                                     unsigned         n __attribute__ ((unused)),
                                     mcfw_memberinfo *members __attribute__ ((unused)),
                                     struct sk_buff  *skb) {
    unsigned char instance_number = (unsigned char) (((unsigned int) privatedata) & 0xff);
    cpmac_instance_t *instance;
    struct net_device *p_dev;

    if(skb == NULL) {
        DEB_ERR("[%s] skb == NULL!\n", __FUNCTION__);
        return;
    }
    assert(instance_number < AVM_CPMAC_MAX_INSTANCES);
    instance = &cpmac_global.instance[instance_number];
#   if defined(CONFIG_ARCH_PUMA5)
    p_dev = cpmac_puma_get_dev();
#   else /*--- #if defined(CONFIG_ARCH_PUMA5) ---*/
    p_dev = instance->dev;
#   endif /*--- #else ---*/ /*--- #if defined(CONFIG_ARCH_PUMA5) ---*/

    if(p_dev == 0) {
        DEB_TEST("[%s] No device for instance %u\n", __FUNCTION__, instance_number);
        dev_kfree_skb_any(skb);
        return;
    }

    /* MC streams go to the LAN */
    skb->uniq_id &= 0xffffff;
    skb->uniq_id |= (CPPHY_PRIO_QUEUE_LAN << 24);
    cpmac_dev_queue_xmit(p_dev, skb, instance_number);
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
int cpmac_mcfw_init(unsigned char instance_number, char *device_name) {
    cpmac_instance_t *instance = &cpmac_global.instance[instance_number];
    int err;

    instance->mcfw_used = 0;
    strncpy(instance->mcfw_netdrv.name, device_name, MCFWDRV_NAMSIZ - 1);
    instance->mcfw_sourceid = 0;
    if(cpmac_global.phy[instance->phy].ports > 1) {
        instance->mcfw_netdrv.netdriver_mc_transmit = cpmac_mc_transmit;
    } else {
        instance->mcfw_netdrv.netdriver_mc_transmit_single = cpmac_mc_transmit_single;
    }
    instance->mcfw_netdrv.privatedata = (void *) (unsigned int) instance_number;

    skb_queue_head_init(&instance->mcfw_queue);
    tasklet_init(&instance->mcfw_tasklet, cpmac_mcfw_handle_tasklet, (unsigned long) instance_number);

    err = mcfw_netdriver_register(&instance->mcfw_netdrv);
    if(err) {
        DEB_WARN("[%s] Could not register mcfw_netdriver!\n", __FUNCTION__);
        return err;
    }
    instance->mcfw_used = 1;
    instance->mcfw_sourceid = mcfw_multicast_forward_alloc_id(instance->mcfw_netdrv.name);
    DEB_INFO("[%s] Register multicast forward for '%s' at instance %u\n", 
             __FUNCTION__, device_name, instance_number);
    return err;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void cpmac_mcfw_exit(unsigned char instance_number) {
    cpmac_instance_t *instance = &cpmac_global.instance[instance_number];

    if(instance->mcfw_used) {
        instance->mcfw_used = 0;
        mcfw_netdriver_unregister(&instance->mcfw_netdrv);
        mcfw_multicast_forward_free_id(instance->mcfw_sourceid);
        tasklet_kill(&instance->mcfw_tasklet);
        DEB_INFO("[%s] done\n", __FUNCTION__);
    }
}

