/*------------------------------------------------------------------------------------------*\
 *   Copyright (C) 2006,2007, ... ,2011 AVM GmbH <fritzbox_info@avm.de>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
\*------------------------------------------------------------------------------------------*/


#include <linux/avm_event.h>
#include "cpmac_debug.h"
#include "cpphy_types.h"
#include "cpphy_mdio.h"
#include "cpmac_ar9_if.h"
#include "cpmac_interface.h"
#include "cpphy_mgmt.h"
#include "mii_reg.h"
#include "cpmac_phys.h"
#if defined(CONFIG_VR9)
#include "ifx_ar9/ifxmips_7port_eth_sw.h"
#include "ifx_ar9/ifxmips_7port_sw_reg.h"
#endif /*--- #if defined(CONFIG_VR9) ---*/

#if defined(CONFIG_AVM_LED_EVENTS) || defined(CONFIG_AVM_LED_EVENTS_MODULE)
#include <linux/avm_led_event.h>
#endif /*--- #if defined(CONFIG_AVM_LED_EVENTS) || defined(CONFIG_AVM_LED_EVENTS_MODULE) ---*/
#if defined(CONFIG_AVM_LED)
#include <linux/avm_led.h>
#endif /*--- #if defined(CONFIG_AVM_LED) ---*/

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void cpmac_phys_init_AR8030(unsigned char phy_number) {
#   if defined(CONFIG_AR9)
    unsigned int value, shift;
    value = (unsigned int) *((volatile unsigned int *) 0xbf20302c);
    shift = (phy_number == 1) ? 9 : 18;
    value &= ~(0x7 << shift);
    value |= 0x5 << shift;
    *((volatile unsigned int *) 0xbf20302c) = (volatile unsigned int) value;
#   endif /*--- #if defined(CONFIG_AR9) ---*/
#   if defined(CONFIG_VR9)
    unsigned short mdio_addr = cpmac_global.phy[phy_number].mdio_address;

    /* Initialize MII configuration in VR9 */
    SW_WRITE_REG32(0x4220 | mdio_addr, PHY_ADDR_0 - (get_mii_nr_from_phy_nr(get_phy_nr_from_net_dev_num(phy_number)) * 4));
    /* Enable status interrupts from the PHY */
    mdio_write(mdio_addr, 0x12, 0x6c20); /* FIXME There should be a better place for this */
#   endif /*--- #if defined(CONFIG_VR9) ---*/
    cpmac_global.power.port_on[cpmac_global.phy[phy_number].port_offset] = 1;
    DEB_INFOTRC("[%s] PHY %u, mdio address %u\n", __FUNCTION__, phy_number, 
                cpmac_global.phy[phy_number].mdio_address);
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void cpmac_phys_init_AR8035(unsigned char phy_number) {
#   if defined(CONFIG_AR9)
    unsigned int value, shift;
    value = (unsigned int) *((volatile unsigned int *) 0xbf20302c);
    shift = (phy_number == 1) ? 9 : 18;
    value &= ~(0x7 << shift);
    value |= 0x5 << shift;
    *((volatile unsigned int *) 0xbf20302c) = (volatile unsigned int) value;
#   endif /*--- #if defined(CONFIG_AR9) ---*/
    cpmac_global.power.port_on[cpmac_global.phy[phy_number].port_offset] = 1;
    DEB_INFOTRC("[%s] PHY %u, mdio address %u\n", __FUNCTION__, phy_number, 
                cpmac_global.phy[phy_number].mdio_address);
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void cpmac_phys_link_status_11G(unsigned char  mdio_address,
                                unsigned char *link,
                                unsigned char *speed,
                                unsigned char *fullduplex) {

    *link = (mdio_read(mdio_address, MII_REG_STATUS) & MII_REG_STATUS_LINK) ? 1 : 0;
    if(*link == 1) {
        unsigned int status = mdio_read(mdio_address, 0x18);
        *speed      = status & 0x3;
        *fullduplex = (status & 0x8) ? 1 : 0;
    }
    DEB_DEBUG("[%s] link 0%x speed 0x%x duplex 0x%x\n", __FUNCTION__, *link, *speed, *fullduplex);

}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void cpmac_phys_link_status_AR8035(unsigned char  mdio_address,
                                   unsigned char *link,
                                   unsigned char *speed,
                                   unsigned char *fullduplex) {
    unsigned int status = 0;

    status = mdio_read(mdio_address, 0x11);
    *link       = (status & (0x1 << 10)) ? 1 : 0;
    *speed      = (status >> 14) & 0x3;
    *fullduplex = (status & (0x1 << 13)) ? 1 : 0;

    DEB_DEBUG("[%s] link 0%x speed 0x%x duplex 0x%x\n", __FUNCTION__, *link, *speed, *fullduplex);
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void cpmac_phys_link_status_vitesse(unsigned char  mdio_address,
                                    unsigned char *link,
                                    unsigned char *speed,
                                    unsigned char *fullduplex) {
    unsigned int status;


    *link = (mdio_read(mdio_address, MII_REG_STATUS) & MII_REG_STATUS_LINK) ? 1 : 0;

    if(*link) {
        status = mdio_read(mdio_address, 0x1c);
        *speed = (status >> 3) & 0x3;
        *fullduplex = (status >> 5) & 0x1;
    }

    DEB_DEBUG("[%s] link 0%x speed 0x%x duplex 0x%x\n", __FUNCTION__, *link, *speed, *fullduplex);
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void cpmac_phys_check_link_PHY(phy_number_t phy_number) {
    cpmac_phy_t *phy = &cpmac_global.phy[phy_number];
    unsigned char port = phy->port_offset;
    enum _avm_event_ethernet_speed speed = avm_event_ethernet_speed_no_link;
    unsigned char link = 0, fullduplex = 0;

    DEB_TRC("[%s] Check link port %u (PHY number %u) %s\n", __FUNCTION__, port, phy_number, cpmac_global.power.port_on[port] ? "poweron":"poweroff");

    if(!cpmac_global.power.port_on[port]) {
        /* Without power, the link information would be wrong */
        DEB_TRC("[%s] Port %u is powered down\n", __FUNCTION__, port);
    } else {
        unsigned char speedvalue = 0xff;
        /* Check real time link information */
        if(phy->f->link_status) {
            phy->f->link_status(phy->mdio_address, &link, &speedvalue, &fullduplex);
        }
        if (link) {     /*--- kein Link kein Speed ---*/
            switch(speedvalue) {
                case 0:  
                    speed = avm_event_ethernet_speed_10M;     
                    break;
                case 1:  
                    speed = avm_event_ethernet_speed_100M;    
                    break;
                case 2:  
                    speed = avm_event_ethernet_speed_1G;      
                    break;
                case 3:
                    speed = avm_event_ethernet_speed_error;      
                    link = 0;
                    break;
                default: 
                    speed = avm_event_ethernet_speed_no_link; 
                    break;
            }
        } else {
            speed = avm_event_ethernet_speed_no_link;
        }

#       if defined(CONFIG_FUSIV_VX180)
        mdio->cpmac_priv->fusiv.set_port_speed(port, speed, fullduplex);
#       endif /*--- #if defined(CONFIG_FUSIV_VX180) ---*/
    }
    cpmac_interface_port_update(port, link, speed, fullduplex);
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void cpmac_phys_init_11G(unsigned char phy_number) {
    cpmac_phy_t *phy = &cpmac_global.phy[phy_number];
    /*--- unsigned int tries = 0; ---*/

    DEB_INFOTRC("[%s] PHY %u, mdio address %u\n", __FUNCTION__, phy_number, phy->mdio_address);
    cpmac_global.power.port_on[phy->port_offset] = 1;
#   if 0
    while(1) {
        if(0xffff != mdio_read(phy->mdio_address, MII_REG_PHY_ID1)) {
            break;
        }
        tries++;
        if(tries >= 3) {
            panic("[%s] Could not revive PHY after %u tries!\n", __FUNCTION__, tries);
        }
        DEB_WARN("[%s] Resetting PHY %u again\n", __FUNCTION__, phy_number);
        cpphy_mgmt_reset_hw(phy->reset_bit, 1);
        ssleep(1);
        cpphy_mgmt_reset_hw(phy->reset_bit, 0);
        ssleep(1);
    }

    mdio_write(phy->mdio_address, MII_REG_CONTROL,   MII_REG_CONTROL_RESET 
                                                   | MII_REG_CONTROL_AUTONEG_ENABLE);
    msleep(500);
    mdio_write(phy->mdio_address, 
               MII_REG_AUTONEG_ADV,
                 MII_REG_AUTONEG_ADV_SF(1)
               | MII_REG_AUTONEG_ADV_10BT_HDX
               | MII_REG_AUTONEG_ADV_10BT_FDX
               | MII_REG_AUTONEG_ADV_100BT_HDX
               | MII_REG_AUTONEG_ADV_100BT_FDX
               | MII_REG_AUTONEG_ADV_PAUSE_SYM
               | MII_REG_AUTONEG_ADV_PAUSE_ASYM);
    /* Declare PHY as part of a multi port device. Set MSPT according to Lantiq hint */
    /* (concerning link problems with GBit switches */
    /* Set MSPT according to Lantiq hint (concerning link problems with GBit switches */
    /* and activate GBit FD and HD for now */
    mdio_write(phy->mdio_address, MII_REG_1000BASET_CTRL,   MII_REG_1000BASET_CTRL_ADV_HD
                                                          | MII_REG_1000BASET_CTRL_ADV_FD
                                                          | MII_REG_1000BASET_CTRL_MSPT);
    /* Enable auto downshift from GBit/s to 100 MBit/s after three autoneg failures */
    mdio_write(cpmac_global.phy[phy_number].mdio_address, 0x14, 0x8006);
    /* Fix against bad links */
    mdio_write(cpmac_global.phy[phy_number].mdio_address, 0x1d, 0x100);
#   endif /*--- #if 0 ---*/
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void cpmac_phys_init_vitesse(unsigned char phy_number) {
    DEB_INFOTRC("[%s] PHY %u\n", __FUNCTION__, phy_number);
    cpmac_global.power.port_on[cpmac_global.phy[phy_number].port_offset] = 1;

#if defined(CONFIG_MACH_FUSIV)
#define GIG_PORT0_BASE   0xb9110880
#define GIG_PORT1_BASE   0xb9150880
#define PORT_CFG_OFF     0x880

    /*--- nicht schön aber selten - fix 10MBit Problem ---*/
    if (port) {
        *(volatile unsigned int *)(GIG_PORT1_BASE + PORT_CFG_OFF) &= ~(0x7<<9);
        *(volatile unsigned int *)(GIG_PORT1_BASE + PORT_CFG_OFF) |=  (0x4<<9);
    } else {
        *(volatile unsigned int *)(GIG_PORT0_BASE + PORT_CFG_OFF) &= ~(0x7<<9);
        *(volatile unsigned int *)(GIG_PORT0_BASE + PORT_CFG_OFF) |=  (0x4<<9);
    }
#endif /*--- #if defined(CONFIG_MACH_FUSIV) ---*/
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
unsigned int cpmac_phys_AR8030_check_irq_status(phy_number_t phy_number) {
    unsigned int status, mask = 0;

    status = mdio_read(cpmac_global.phy[phy_number].mdio_address, 0x13);
    if(status) {
        mask = mdio_read(cpmac_global.phy[phy_number].mdio_address, 0x12);
    }
#   if (CPMAC_DEBUG_LEVEL & CPMAC_DEBUG_LEVEL_DEBUG)
    DEB_DEBUG("[%s] PHY %u: IRQ %sset (status %#x, mask %#x)\n", __FUNCTION__, phy_number, (status & mask) ? "" : "not ", status, mask); 
#   else /*--- #if (CPMAC_DEBUG_LEVEL & CPMAC_DEBUG_LEVEL_DEBUG) ---*/
    if(status & mask) {
        DEB_TRC("[%s] PHY %u: IRQ %sset (status %#x, mask %#x)\n", __FUNCTION__, phy_number, (status & mask) ? "" : "not ", status, mask); 
    }
#   endif /*--- #else ---*/ /*--- #if (CPMAC_DEBUG_LEVEL & CPMAC_DEBUG_LEVEL_DEBUG) ---*/
    return status & mask;
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
unsigned int cpmac_phys_11g_check_irq_status(phy_number_t phy_number) {
    unsigned int status, mask = 0;

    status = mdio_read(cpmac_global.phy[phy_number].mdio_address, 0x1A);
    if(status) {
        mask = mdio_read(cpmac_global.phy[phy_number].mdio_address, 0x19);
    }
#   if (CPMAC_DEBUG_LEVEL & CPMAC_DEBUG_LEVEL_DEBUG)
    DEB_DEBUG("[%s] PHY %u: IRQ %sset (status %#x, mask %#x)\n", __FUNCTION__, phy_number, (status & mask) ? "" : "not ", status, mask); 
#   else /*--- #if (CPMAC_DEBUG_LEVEL & CPMAC_DEBUG_LEVEL_DEBUG) ---*/
    if(status & mask) {
        DEB_TRC("[%s] PHY %u: IRQ %sset (status %#x, mask %#x)\n", __FUNCTION__, phy_number, (status & mask) ? "" : "not ", status, mask); 
        DEB_TRC("[%s] %s%s%s%s%s%s%s%s%s%s%s%s", __FUNCTION__, 
                (1 << 0)  & status ? " Link state change\n" : "",
                (1 << 1)  & status ? " Link speed change\n" : "",
                (1 << 2)  & status ? " Duplex Mode change\n" : "",
                (1 << 3)  & status ? " MDIX Change\n" : "",
                (1 << 4)  & status ? " MDI Polarity Change\n" : "",
                (1 << 5)  & status ? " Link Speed Auto Down Shift Detect\n" : "",
                (1 << 10) & status ? " Auto Negotiation Complete\n" : "",
                (1 << 11) & status ? " Auto Negotioation Error\n" : "",
                (1 << 12) & status ? " Next Page Transmitter\n" : "",
                (1 << 13) & status ? " Next Page Received\n" : "",
                (1 << 14) & status ? " Master-Slave Resolution Error\n" : "",
                (1 << 15) & status ? " Wake On Lan Event\n" : "\n");
    }
#   endif /*--- #else ---*/ /*--- #if (CPMAC_DEBUG_LEVEL & CPMAC_DEBUG_LEVEL_DEBUG) ---*/
    return status & mask;
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
unsigned long cpmac_phys_poll_irq_status(void) {
    phy_number_t phy_number;

    for(phy_number = 0; phy_number < AVM_CPMAC_MAX_PHYS; phy_number++) {
        if(   cpmac_global.phy[phy_number].use_irq_poll
           && cpmac_global.phy[phy_number].f->check_irq_status) {
           if(cpmac_global.phy[phy_number].f->check_irq_status(phy_number)) {
               cpphy_mgmt_work_schedule(CPMAC_WORK_LINK_CHECK, 0);
           }
        }
    }
    return CPMAC_LINK_TICK_LENGTH;
}


