/*------------------------------------------------------------------------------------------*\
 *   Copyright (C) 2006,2007, ... ,2011 AVM GmbH <fritzbox_info@avm.de>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
\*------------------------------------------------------------------------------------------*/


#if defined(CONFIG_MACH_AR7240) || defined(CONFIG_MACH_AR724x)

#include <linux/delay.h>
#include <linux/wait.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/sched.h>
#include "cpmac_if.h"
#include "cpmac_main.h"
#include "cpmac_virian_if.h"
#include "cpphy_types.h"
#include "cpmac_debug.h"
#include "cpphy_mdio.h"
#include "cpphy_if.h"
#include "cpmac_eth.h"
#include <asm/mach_avm.h>
#include "cpphy_mgmt.h"
/*--- #include <mach/hw_gpio.h> ---*/

/* TODO list */
/* - Get MAC address from environment and assign it to the PHY (from cpphy_if_init) */
 

/*------------------------------------------------------------------------------------------*\
 * Global resources
\*------------------------------------------------------------------------------------------*/
void *virian_phy_dev = NULL;
extern cpmac_global_t cpmac_global;


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void cpphy_mdio_user_access_write(unsigned short regaddr,
                                  unsigned short phyaddr,
                                  unsigned short data) {
    ag7240_mdio_write(phyaddr, regaddr, data);
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
unsigned int cpphy_mdio_user_access_read(unsigned short regaddr, unsigned short phyaddr) {
    return ag7240_mdio_read(phyaddr, regaddr);
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static void cpmac_virian_start(void) {
    cpphy_mgmt_global_power_set(CPPHY_POWER_GLOBAL_ON);
    cpphy_mgmt_work_schedule(CPMAC_WORK_TICK, 0);
    (void) wait_event_interruptible(cpmac_global.power.waitqueue, (cpmac_global.power.global == CPPHY_POWER_GLOBAL_NONE));
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void cpmac_virian_set_dev(struct net_device *dev) {
    cpmac_priv_t *cpmac_priv;
    DEB_INFOTRC("[%s]\n", __FUNCTION__);

    cpmac_priv = kzalloc(sizeof(cpmac_priv_t), GFP_KERNEL);
    assert(cpmac_priv); /* Can not work without memory */
    cpmac_global.instance[0].cpmac_priv = cpmac_priv;
    cpmac_global.instance[0].mdio.cpmac_priv = cpmac_priv;
    cpmac_global.instance[0].dev = dev;
    virian_phy_dev = dev;
    if(cpmac_global.instance[0].is_open == 1) {
        DEB_TEST("[%s] start cpmac\n", __FUNCTION__); /* FIXME */
        cpmac_virian_start();
    }
}
EXPORT_SYMBOL(cpmac_virian_set_dev);


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void cpmac_virian_open(struct net_device *dev) {
    DEB_INFOTRC("[%s]\n", __FUNCTION__);

    /* FIXME Support more than one instance/device */
    cpmac_global.instance[0].dev = dev;

    cpphy_mgmt_global_power_set(CPPHY_POWER_GLOBAL_ON);
    cpphy_mgmt_work_schedule(CPMAC_WORK_TICK, 0);
}
EXPORT_SYMBOL(cpmac_virian_open);


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
int cpmac_virian_ioctl(struct net_device *dev __attribute__ ((unused)),
                       struct ifreq *ifr,
                       int cmd __attribute__ ((unused))) {
    struct avm_cpmac_ioctl_struct ioctl_struct;
    cpmac_err_t err;

    if(copy_from_user(&ioctl_struct, ifr->ifr_data, sizeof(ioctl_struct))) {
        DEB_ERR("[%s] copy from user failed\n", __FUNCTION__);
        return -EFAULT;
    }
    err = cpphy_switch_ioctl(&cpmac_global.instance[0].mdio, &ioctl_struct);
    if(err != CPMAC_ERR_NOERR) {
        return -EFAULT;
    }
    if(copy_to_user(ifr->ifr_data, &ioctl_struct, sizeof(ioctl_struct))) {
        DEB_ERR("[%s] copy to user failed\n", __FUNCTION__);
        return -EFAULT;
    }
    return 0;
}
EXPORT_SYMBOL(cpmac_virian_ioctl);

/*------------------------------------------------------------------------------------------*\
 * TODO: The receive functions here and in cpmac_if.c should be unified
\*------------------------------------------------------------------------------------------*/
void cpmac_virian_if_rx(struct sk_buff *skb) {
#   if defined(CONFIG_IP_MULTICAST_FASTFORWARD)
    int hlen = skb->data - skb_mac_header(skb);
#   endif /*--- #if defined(CONFIG_IP_MULTICAST_FASTFORWARD) ---*/

#   if defined(CONFIG_IP_MULTICAST_FASTFORWARD)
    skb_push(skb, hlen);
    mcfw_snoop_recv(&cpmac_global.instance[0].mcfw_netdrv, 0, skb);
    skb_pull(skb, hlen);
#   endif /*--- #if defined(CONFIG_IP_MULTICAST_FASTFORWARD) ---*/
    netif_receive_skb(skb);
}
EXPORT_SYMBOL(cpmac_virian_if_rx);


#endif /*--- #if defined(CONFIG_MACH_AR7240) || defined(CONFIG_MACH_AR724x) ---*/

