/**
** FILE NAME    : ifxmips_3port_eth_sw.h
** PROJECT      : IFX UEIP
** MODULES      : ETH
** DATE         : 24 July 2009
** AUTHOR       : Reddy Mallikarjuna
** DESCRIPTION  : IFX ETH driver header file
** COPYRIGHT    :       Copyright (c) 2009
**                      Infineon Technologies AG
**                      Am Campeon 1-12, 85579 Neubiberg, Germany
**
**    This program is free software; you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation; either version 2 of the License, or
**    (at your option) any later version.
**
** HISTORY
** $Date            $Author         $Comment
** 24 July 2009     Reddy Mallikarjuna  Initial release
*******************************************************************************/

#ifndef _IFXMIPS_3PORT_SW_H_
#define _IFXMIPS_3PORT_SW_H_
/*!
  \file ifxmips_3port_eth_sw.h
  \ingroup IFX_ETH_DRV
  \brief Header file for ARx ETH driver internal definitions
*/

#include <ifx_gpio.h>

/**  default board related configuration  */
/* MII0 mode */
#if defined(CONFIG_MII0_RGMII_MODE)
#define MII0_MODE_SETUP                         RGMII_MODE
#else
#define MII0_MODE_SETUP                         RGMII_MODE_100MB
#endif

#define MII1_GPHY_INT_EXT_INT_NO              3

#if defined(CONFIG_MII1_RGMII_MODE_100MB)
  #define MII1_MODE_SETUP                       RGMII_MODE_100MB
#elif defined(CONFIG_MII1_RGMII_MODE)
  #define MII1_MODE_SETUP                       RGMII_MODE
#elif defined(CONFIG_MII1_AUTONEG)
  #define MII1_MODE_SETUP                       RGMII_MODE_100MB  //  default setting before auto-negotiate
  #define MII1_GPHY_INT                         IFX_EIU_IR3           
#else
  #error you must specify MII1 link state in current stage
#endif

#ifdef CONFIG_SW_ROUTING_MODE
    #define CONFIG_PMAC_DMA_ENABLE          1   /*g_pmac_dma */
    #define CONFIG_DMA_PMAC_ENABLE          1   /*g_dma_pmac*/
    #define INGRESS_DIRECT_FORWARD_ENABLE   1
#else
    #define CONFIG_PMAC_DMA_ENABLE          0   /*g_pmac_dma */
    #define CONFIG_DMA_PMAC_ENABLE          0   /*g_dma_pmac*/
    #define INGRESS_DIRECT_FORWARD_ENABLE   0
#endif

#if defined(CONFIG_PMAC_DMA_ENABLE) && CONFIG_PMAC_DMA_ENABLE
    #define NUM_ETH_INF                     2
#else
    #define NUM_ETH_INF                     1
#endif
#define CONFIG_IFX_MAC1                     1
#define ENABLE_SWITCH_FLOW_CONTROL          1
/* select 25MHz clock based on design board*/
#define SELECT_25MHZ_CLOCK_FROM_GPIO        0
/** GPIO 42 (P2.10) */
#define GPIO_MDIO                           42
/** GPIO 43 (P2.11) */
#define GPIO_MDC                            43
 /*GPIO 3 (P0.3) */
#define GPIO_CLK_OUT2                       3
/*--- AVM HW ---*/
#define GPIO_RESET_GPHY0                    34
#define GPIO_RESET_GPHY1                    36
/** HW related configuration */
#define PHY0_ADDR                           0x0
#define PHY1_ADDR                           0x1

#define TANTOS_CHIP_ID                      0x2599
#define PHY11G_CHIP_ID                      0x0302
#define PHY11G_SUB_ID                       0x60D1
#define ATH8030_CHIP_ID                     0x004D
#define ATH8030_SUB_ID                      0xD000

#define NO_PHY_ID                           0xFFFF

/*!
  \addtogroup IFX_ETH_DRV_STRUCTURE
 */
 
/*@{*/

/*! \typedef  cpu_ingress_pkt_header_t
* \brief Pseudo MAC(PMAC) cpu ingress packet header structure(4-bytes)

    This structure is used internal purpose
*/
typedef struct
{
    u32	res31_27                :5; /*!< reserved bits */
    u32 	SPID                :3; /*!< Source port ID*/
    u32	res23_18                :6; /*!< reserved bits */
    u32     DPID                :2; /*!< Destination Physical Port ID valid only when DIRECT is set*/
    u32	res15_10                :6; /*!< reserved bits */
    u32	QID                     :2; /*!< Queue ID of destination Physical port, valid only when Direct is set*/
    u32	res7_1                  :7; /*!< reserved bits */
    u32      DIRECT             :1; /*!< Direct Forwarding Enable*/
}cpu_ingress_pkt_header_t;

/*! \typedef  cpu_egress_pkt_header_t
* \brief Pseudo MAC(PMAC) cpu egress packet header structure (8-bytes)

    This structure is used internal purpose
*/
typedef struct
{
    u32 ipOffset                :8; /*!< IP offset*/
    u32 destPortMap             :8; /*!< destination port map*/
    u32 srcPortID               :3; /*!< Source port id*/
    u32 inputTag                :1; /*!< input tag flag, packet is received with VALN tag or not*/
    u32 managementPacket        :1; /*!< management flag, packet is classifield as managemnt packet*/
    u32 outputTag               :1; /*!< packet is transmitted with VLAN tag to CPU */
    u32 queueID                 :2; /*!< Queue ID*/
    u32 crcGenerated            :1; /*!< packet include CRC*/
    u32 crcErr                  :1; /*!< crc error packet*/
    u32 arpRarpPacket           :1; /*!< ARP/RARP packet*/
    u32 PPPoEPacket             :1; /*!< PPPoE packet*/
    u32 iPv6Packet              :1; /*!< IPv6 packet*/
    u32 res                     :1; /*!< reserved */
    u32 iPv4Packet              :1; /*!< IPv4 packet*/
    u32 mirrored                :1; /*!< Packet was mirrored*/
    u32 res_31_16               :16;    /*!< reserved */
    u32 PKTLEN_8                :8; /*!< packet length Lower 8 bits*/
    u32 res_7_3                 :5; /*!< reserved bits 7~3*/
    u32 PKTLEN_3                :3; /*!< packet length higher 3 bits*/
}cpu_egress_pkt_header_t;
/* @} */


#endif /*_IFXMIPS_3PORT_SW_H_  */

