/**
** FILE NAME    : ifxmips_7port_eth_sw.h
** PROJECT      : IFX UEIP
** MODULES      : ETH
** DATE         : 24 July 2009
** AUTHOR       : Reddy Mallikarjuna
** DESCRIPTION  : IFX ETH driver header file
** COPYRIGHT    :       Copyright (c) 2009
**                      Infineon Technologies AG
**                      Am Campeon 1-12, 85579 Neubiberg, Germany
**
**    This program is free software; you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation; either version 2 of the License, or
**    (at your option) any later version.
**
** HISTORY
** $Date            $Author         $Comment
** 24 July 2009     Reddy Mallikarjuna  Initial release
*******************************************************************************/

#ifndef _IFXMIPS_7PORT_ETH_SW_H_
#define _IFXMIPS_7PORT_ETH_SW_H_

/*!
  \file ifxmips_7port_eth_sw.h
  \ingroup IFX_ETH_DRV
  \brief Header file for VRx ETH driver internal definitions
*/

#include <ifx_gpio.h>

/**  default board related configuration  */
#if defined(CONFIG_MII0_PORT_ENABLED)
#define CONFIG_MAC0                         1
#else
#define CONFIG_MAC0                         0
#endif

#if defined(CONFIG_MII1_PORT_ENABLED)
#define CONFIG_MAC1                         1
#else
#define CONFIG_MAC1                         0
#endif

#if defined(CONFIG_MII2_PORT_ENABLED)
#define CONFIG_MAC2                         1
#else
#define CONFIG_MAC2                         0
#endif

#if defined(CONFIG_MII3_PORT_ENABLED)
#define CONFIG_MAC3                         1
#else
#define CONFIG_MAC3                         0
#endif

#if defined(CONFIG_MII4_PORT_ENABLED)
#define CONFIG_MAC4                         1
#else
#define CONFIG_MAC4                         0
#endif

#if defined(CONFIG_MII5_PORT_ENABLED)
#define CONFIG_MAC5                         1
#else
#define CONFIG_MAC5                         0
#endif

#if defined(CONFIG_MII0_RGMII_MAC_MODE)
#define MII0_MODE_SETUP                 RGMII_MODE
#elif defined(CONFIG_MII0_RMII_MAC_MODE)
#define MII0_MODE_SETUP                 RMII_MAC_MODE
#elif defined(CONFIG_MII0_RMII_PHY_MODE)
#define MII0_MODE_SETUP                 RMII_PHY_MODE
#elif defined(CONFIG_MII0_MII_MAC_MODE)
#define MII0_MODE_SETUP                 MII_MAC_MODE
#elif defined(CONFIG_MII0_MII_PHY_MODE)
#define MII0_MODE_SETUP                 MII_PHY_MODE
#endif

#if defined(CONFIG_MII1_RGMII_MAC_MODE)
#define MII1_MODE_SETUP                 RGMII_MODE
#elif defined(CONFIG_MII1_RMII_MAC_MODE)
#define MII1_MODE_SETUP                 RMII_MAC_MODE
#elif defined(CONFIG_MII1_RMII_PHY_MODE)
#define MII1_MODE_SETUP                 RMII_PHY_MODE
#elif defined(CONFIG_MII1_MII_MAC_MODE)
#define MII1_MODE_SETUP                 MII_MAC_MODE
#elif defined(CONFIG_MII1_MII_PHY_MODE)
#define MII1_MODE_SETUP                 MII_PHY_MODE
#endif

#if defined(CONFIG_MII2_GMII_MODE)
#define MII2_MODE_SETUP                 GMII_MAC_MODE
#elif defined(CONFIG_MII2_MII_MAC_MODE)
#define MII2_MODE_SETUP                 MII_MAC_MODE
#endif

#if defined(CONFIG_MII3_MII_MAC_MODE)
#define MII3_MODE_SETUP                 MII_MAC_MODE
#endif

#if defined(CONFIG_MII4_GMII_MODE)
#define MII4_MODE_SETUP                 GMII_MAC_MODE
#elif defined(CONFIG_MII4_MII_MAC_MODE)
#define MII4_MODE_SETUP                 MII_MAC_MODE
#endif

#if defined(CONFIG_MII5_RGMII_MODE)
#define MII5_MODE_SETUP                 RGMII_MODE
#elif defined(CONFIG_MII5_MII_MAC_MODE)
#define MII5_MODE_SETUP                 MII_MAC_MODE
#endif


#ifdef CONFIG_VR9
    #define NUM_ETH_INF                     4
    #define CONFIG_PMAC_DMA_ENABLE          1   /*g_pmac_dma */
    #define CONFIG_DMA_PMAC_ENABLE          1   /*g_dma_pmac*/
#else /*--- #ifdef CONFIG_VR9 ---*/
#ifdef CONFIG_SW_ROUTING_MODE
    #define CONFIG_PMAC_DMA_ENABLE          1   /*g_pmac_dma */
    #define CONFIG_DMA_PMAC_ENABLE          1   /*g_dma_pmac*/
#else
    #define CONFIG_PMAC_DMA_ENABLE          0   /*g_pmac_dma */
    #define CONFIG_DMA_PMAC_ENABLE          0   /*g_dma_pmac*/
#endif

#if defined(CONFIG_PMAC_DMA_ENABLE) && CONFIG_PMAC_DMA_ENABLE
    #define NUM_ETH_INF                     2
#else
    #define NUM_ETH_INF                     1
#endif
#endif /*--- #else ---*/ /*--- #ifdef CONFIG_VR9 ---*/

#if defined(LITTLE_ENDIAN) || defined(BIG_ENDIAN) || defined(__LITTLE_ENDIAN) || defined(__BIG_ENDIAN)

/*!
  \addtogroup IFX_ETH_DRV_STRUCTURE
 */
 
/*@{*/

/*! \typedef  cpu_ingress_pkt_header_t
* \brief Pseudo MAC(PMAC) cpu ingress packet header structure(4-bytes)

    This structure is used internal purpose
*/
typedef struct /* oder sw_eg_pkt_header */
{
    u32 resv_7_3            :5; /*!< Reserved bits */
    u32 SPID                :3; /*!< Source Logical Port ID */
    u32 CRCGEN_DIS          :1; /*!< CRC Generation Disable */
    u32 resv_14_11          :4;  /*!< Reserved bits */
    u32 DPID                :3; /*!< Destination Group and Port ID */
    u32 Port_Map_EN         :1; /*!< Destination Ethernet Port Map Enable */
    u32 Port_Map_Sel        :1; /*!< Destination Ethernet Port Map or Mask */
    u32 LRN_DIS             :1; /*!< Source MAC Address Learning Disable */
    u32 CLASS_EN            :1; /*!< Destination Class Enable */
    u32 Ig_Tr_CLASS         :4; /*!< Target Traffic Class */
    u32 resv_31             :1; /*!< Reserved bit*/
    u32 Dest_Port_Map       :6; /*!< Destination Ethernet Port Map */
    u32 DPID_EN             :1; /*!< Destination Group ID Enable */
} cpu_ingress_pkt_header_t;

/*! \typedef  cpu_egress_pkt_header_t
* \brief Pseudo MAC(PMAC) cpu egress packet header structure (8-bytes)

    This structure is used internal purpose
*/
typedef struct
{
    u32 b0_resv_7_6         :2;  /*!< Reserved bits*/
    u32 IPOFF               :6;  /*!< IP Offset */
    u32 PORTMAP             :8; /*!< Destination Group and Port Map */
    u32 SLPID               :3; /*!< Source Logical Port ID */
    u32 b2_reserved_4_3     :2; /*!< Reserved bits*/
    u32 IS_TAG              :1; /*!< Is Tagged */
    u32 b2_reserved_1_0     :2; /*!< Reserved bits*/
    u32 b3_reserved_7_4     :4; /*!< Reserved bits*/
    u32 PPPoES              :1; /*!< PPPoE Session Packet */
    u32 IPv6                :1; /*!< IPv6 Packet */
    u32 IPv4                :1; /*!< IPv4 Packet */
    u32 MRR                 :1; /*!< Mirrored */
    u32 b4_resv_7_6         :2; /*!< Reserved bits*/
    u32 PKT_LEN_HI          :6; /* Packet Length High Bits */
    u32 PKT_LEN_LO          :8; /* Packet Length Low Bits */
    u32 b6_resv_7_0         :8; /*!< Reserved bits*/
    u32 b7_resv_7           :1; /*!< Reserved bits*/
    u32 SPPID               :3; /*!< Source Physical Port ID */
    u32 Eg_Tr_CLASS         :4; /*!< Traffic Class */
} cpu_egress_pkt_header_t;
/* @} */
#endif /*--- #if defined(LITTLE_ENDIAN) || defined(BIG_ENDIAN) || defined(__LITTLE_ENDIAN) || defined(__BIG_ENDIAN) ---*/

unsigned int get_mii_nr_from_phy_nr(unsigned int phy_nr);
unsigned int get_phy_nr_from_net_dev_num(unsigned int net_dev_number);
int ifx_eth_module_init(void);
void ifx_eth_module_exit(void);
int MAC_TABLE_PRINT(void);

#endif /*_IFXMIPS_7PORT_ETH_SW_H_  */
